#!/usr/bin/env python

import sys
import random

import shelve
import atexit

import os.path
import glob

import pickle

import pygame.image
import pygame.font
import pygame.key
import pygame.mouse
import pygame.display
import pygame.mixer
import pygame.mixer_music
import pygame.time
import pygame.event
import pygame.mouse
import pygame

from pygame.locals import *

###############################
#                             #
# Graphics handling functions #
#                             #
###############################

# load an image with transparency (key #ff00ff)
# convert to display surface format
def makeImage(filename):
	image = pygame.image.load(filename)
	image.set_colorkey((255,0,255),RLEACCEL)
	image.convert()
	return image

# load a whole set of ground tiles from a manifest file
def makeGroundTiles(filename):
	f = open(os.path.join('assets',filename + '.tiles'))
	tiles = []
	for line in f.readlines():
		line = line.rstrip()
		if line[0] == '#':
			continue
		tiles.append(makeImage(os.path.join('gfx','tiles',line)))
	f.close()
	return tiles

# load a whole object tile set
def makeObjectTileSet(manifest):
	tiles = []
	# read the tile definition file
	f = open(os.path.join('assets',manifest + '.obj'))
	for line in f.readlines():
		line = line.rstrip()
		if line[0] == '#':
			continue
		tiles.append(makeImage(os.path.join('gfx','objects',line)))
	f.close()
	return tiles

def drawBackground(level, tileset, screen):
	# offset of *top-left corner* of top left tile
	# 400 - 19 (half tile width)
	# 120 - 9 (half height of top face of isometric cube)
	rowstartx = 381
	rowstarty = 111
	for i in level:
		xoff = rowstartx
		yoff = rowstarty
		rowstartx -= 18
		rowstarty += 9
		for j in i:
			screen.blit(tileset[j],(xoff,yoff))
			xoff += 18
			yoff += 9

# render overlay (flat tiles - roads, rails, etc. - not objects!)
# pass in the overlay map, the collection of overlay tilesets, and the surface to draw on
def drawOverlay(level, tilesets, surface):
	iy = 0
	# offset of *top-left corner* of top left tile
	# 400 - 19 (half of overlay tile width), 
	# 120 - 8 (half height of top face of isometric cube)
	xstart = 381
	ystart = 112
	for i in level:
		xoff = xstart - (iy * 18)
		yoff = ystart + (iy * 9)
		for j in i:
			if not j == -1:
				surface.blit(tilesets[j[0]].tiles[j[1]].image,(xoff,yoff))
			xoff += 18
			yoff += 9
		iy += 1

# render objects & mobs (non-flat tiles; can be any height, but only 1 wide)
# objects must be considered in terms of their *bottom-left corner* because of the varying heights
# the "hotspot" of a mob - the centre of its bottom face - moves in two steps:
# 	- from source-tile centre to joining edge of source & dest
#	- from joining edge of source & dest to centre of dest
def drawObjectsAndMobs(level, tileset, mobs, surface, frameno, targets = None):
	ix = 0
	iy = 0
	# offset of *bottom-left* of top left tile
	# 400 - 19 (half of overlay tile width), 
	# 120 + 9 (half bottom of top face of isometric cube)
	xstart = 381
	ystart = 129
	for i in level:
		xoff = xstart - (iy * 18)
		yoff = ystart + (iy * 9)
		for j in i:
			if not j == -1:
				surface.blit(tileset[j],(xoff,yoff-tileset[j].get_height()))
			xoff += 18
			yoff += 9
			for mob in mobs:
				if mob.location[0] == ix and mob.location[1] == iy:
					mob.draw(surface, frameno, targets)
			ix += 1
		ix = 0
		iy += 1

###################
#                 #
# Path tiles code #
#                 #
###################

# 'record' class for path tiles
class pathTile:
	pass

# 'container' class for a whole tile set
class pathTileSet:
	def __init__(self, manifest, number):
		self.joinedTo = (None, None)
		self.tiles = []
		self.number = number
		# read the tile definition file
		f = open(os.path.join('assets',manifest + '.paths'))
		i = 0
		for line in f.readlines():
			line = line.rstrip()
			if line[0] == '#':
				continue
			if i == 0:
				i = 1
				# first, find the tile graphic
				newTile = pathTile()
				newTile.image = makeImage(os.path.join('gfx','overlays',line))
				newTile.routingTable = []
				newTile.parent = number
				newTile.index = len(self.tiles)
				self.tiles.append(newTile)
			elif i == 1:
				# next, build the routing table - source, destination
				#
				# source/dest map:
				#  |0|
				# -+-+-
				# 1| |2
				# -+-+-
				#  |3|
				#
				if not line == 'end':
					self.tiles[len(self.tiles)-1].routingTable.append(line.split(','))
				else:
					i = 0
		f.close()

####################
#                  #
# Mobile unit code #
#                  #
####################

# parse a .mob file and return a tileset
def loadMobTiles(manifest):
	# read the unit definition file
	f = open(os.path.join('assets',manifest + '.mob'), 'r')
	i = 0
	for line in f.readlines():
		line = line.rstrip()
		# ignore comments
		if line[0] == '#':
			continue
		if i == 0:
			i = 1
			# first, find the list of tiles for each direction
			tiles = line.split(',')
			tileset = []
			for tile in tiles:
				tileset.append(makeImage(os.path.join('gfx','mobs',tile)))
	f.close()
	return tileset

# map destinations from source's perspective to sources from destination's perspective
sourceMap = { '0':'3', '3':'0', '1':'2', '2':'1' }

# mobile unit class
class mob:
	# map source/destination pairs to tileset indices
	# directions are:
	# 0 -left, 1 - leftdown, 2 - down, 3 - downright, 4 - right, 5 - rightup, 6 - up, 7 - upleft,
	# 11 - rightdown, 9 - downleft, 13 - upright, 15 - leftup
	# difference between the two sets of diagonals is whether the path goes through the (upper|left-hand)/(lower|right-hand) side of the tile
	# directions above 7 are the variations on diagonals - subtract 8 to get tile index
	directionMap = { '21': 0, '23': 1, '03': 2, '02': 3, '12': 4, '10': 5, '30': 6, '31': 7, \
		'13': 11, '01': 9, '32': 13, '20': 15}
	# starting offsets for the various directions (relative to bottom left)
	startingPoints = { 0: (8,4), 1: (8,4), 2: (8, -4), 3: (9, -4), 4: (-8, -4), 5: (-8, -5), 6: (-8, 4), 7: (-9, 4), \
		11: (-9, -4), 9: (8, -5), 13: (-8, 4), 15: (9, 4) }
	# x delta, y delta and journey length for the various directions
	graphicalDeltas = { 0: (-2, -1, 9), 1: (-2, 0, 9), 2: (-2, 1, 9), 3: (0, 1, 9), 4: (2, 1, 9), 5: (2, 0, 9), 6: (2, -1, 9), 7: (0, -1, 9), \
		11: (0, 1, 9), 9: (-2, 0, 9), 13: (2, 0, 9), 15: (0, -1, 9) }
	gridDeltas = { 0: (-1, 0), 1: (0, 1), 2: (0, 1), 3: (1, 0), 4: (1, 0), 5: (0, -1), 6: (0, -1), 7: (-1, 0), \
		11: (0, 1), 9: (-1, 0), 13: (1, 0), 15: (0, -1) }
	# pass in tileset returned from loadMobTiles, pair of origin coordinates, overlay map and path sets
	def __init__(self, tileset, location, level, pathsets):
		# our graphics & the overlay set
		self.tiles = tileset
		self.level = level
		# tile we're currently in/moving through
		self.location = []
		self.location.append(location[0])
		self.location.append(location[1])
		# path set on which we travel
		self.pathSet = pathsets[self.level[self.location[1]][self.location[0]][0]]
		# direction we're going in - decide this based on our start tile
		self.level = level
		numRoutes = len(self.pathSet.tiles[self.level[self.location[1]][self.location[0]][1]].routingTable)
		myRoute = random.choice(range(numRoutes))
		self.source = self.pathSet.tiles[self.level[self.location[1]][self.location[0]][1]].routingTable[myRoute][0]
		self.dest = self.pathSet.tiles[self.level[self.location[1]][self.location[0]][1]].routingTable[myRoute][1]
		self.direction = mob.directionMap[str(self.source) + str(self.dest)]
		# progress from one side to the next
		self.progress = 0
		# we are active
		self.valid = True
		# we haven't been drawn yet
		self.frameno = -1
		# and we haven't been through any target squares
		self.targetsHit = None
	# draw self
	def draw(self, surface, fno, targets = None):
		if self.frameno == fno:
			return
		self.frameno = fno
		# offset of *bottom-left corner* of top left tile
		# 400 - 19 (half of overlay tile width), 
		# 120 + 9 (half bottom of top face of isometric cube)
		x = 381
		y = 129
		# move to actual tile itself
		x += (self.location[0] * 18) - (self.location[1] * 18)
		y += (self.location[1] * 9) + (self.location[0] * 9)
		# find tile - directions above 7 should have 8 subtracted to find the graphic
		index = self.direction
		if index > 7:
			index -= 8
		# add starting points
		x += mob.startingPoints[self.direction][0]
		y += mob.startingPoints[self.direction][1]
		# add deltas
		x += mob.graphicalDeltas[self.direction][0] * self.progress
		y += mob.graphicalDeltas[self.direction][1] * self.progress
		# animation progress
		self.progress += 1
		if (self.progress == mob.graphicalDeltas[self.direction][2]):
			self.progress = 0
			# move to adjoining grid square
			self.location[0] += mob.gridDeltas[self.direction][0]
			self.location[1] += mob.gridDeltas[self.direction][1]
			# if we can't travel on to the adjoining square, say so
			self.valid = False
			# there may be no track on the adjoining square
			if self.level[self.location[1]][self.location[0]] != -1:
				# check new location to determine next direction
				for route in self.pathSet.tiles[self.level[self.location[1]][self.location[0]][1]].routingTable:
					if sourceMap[self.dest] == route[0]:
						self.source = route[0]
						self.dest = route[1]
						self.direction = mob.directionMap[str(self.source) + str(self.dest)]
						self.valid = True
						# if we have just moved to a target square (that we haven't already passed),
						# then add it to our list
						#print "train is checking "+str(self.location)+" against "+str(targets)
						if targets != None and (self.location in targets) and ((self.targetsHit == None) or (self.location not in self.targetsHit)):
							#print "been through "+str(self.location)
							if self.targetsHit == None:
								self.targetsHit = []
							self.targetsHit.append([self.location[0],self.location[1]])
		# perform the blit - the position passed to blit() is *upper* left, so subtract object height
		if self.valid:
			y -= self.tiles[index].get_height()
			surface.blit(self.tiles[index], (x, y))

#############
#           #
# Menu code #
#           #
#############

# open the settings shelf
settings = shelve.open(os.path.join(os.path.expanduser('~'),'.traintris_settings'))
# default settings & scores
if len(settings.keys()) == 0:
	settings['sound'] = True
	settings['music'] = True
	settings['samplerate'] = 44100
	settings['buffer'] = 2048
	settings['fullscreen'] = False
# open the hiscores pickle
try:
	scores = pickle.load(open(os.path.join(os.path.expanduser('~'),'.traintris_scores'),'r'))
except:
	scores = []
	for i in range(10):
		scores.append(['traintris',(10-i)*100])

def exitFunc():
	print "Bye!"
	global scores
	global settings
	pickle.dump(scores,open(os.path.join(os.path.expanduser('~'),'.traintris_scores'),'w'))
	settings.close()

atexit.register(exitFunc)

# menu funcs
def TrueFalseStr(v):
	if v:
		return 'True'
	else:
		return 'False'
def OnOffStr(v):
	if v:
		return 'On'
	else:
		return 'Off'
def doFullscreen(left):
	global settings
	if left:
		settings['fullscreen'] = False
	else:
		settings['fullscreen'] = True
def doMusic(left):
	global settings
	if left:
		settings['music'] = False
	else:
		settings['music'] = True
def doSound(left):
	global settings
	if left:
		settings['sound'] = False
	else:
		settings['sound'] = True
def doSamplerate(left):
	global settings
	if left and settings['samplerate'] > 11025:
		settings['samplerate'] = settings['samplerate'] / 2
	elif (not left) and settings['samplerate'] < 44100:
		settings['samplerate'] = settings['samplerate'] * 2
def doBuffer(left):
	global settings
	if left and settings['buffer'] > 32:
		settings['buffer'] = settings['buffer'] / 2
	elif (not left) and settings['buffer'] < 65536:
		settings['buffer'] = settings['buffer'] * 2
def startEditor():
	global nextLoop
	nextLoop = levelEditor()
def startGame():
	global nextLoop
	nextLoop = mainGame()
def startCustomGame():
	global nextLoop
	nextLoop = enterLevelName()

# make the main menu structure
def getMainMenu():
	global settings
	mainMenu = [
		[
			['New Game', 'High Scores', 'Options', 'Level Editor', 'Quit'],
			[2, 3, 1, startEditor, sys.exit],
			[None, None, None, None]
		],
		[
			['Full Screen: '+TrueFalseStr(settings['fullscreen']),'Music: '+OnOffStr(settings['music']),'Sound: '+OnOffStr(settings['sound']), \
				'Sound buffer size: '+str(settings['buffer']), 'Sample rate: '+str(settings['samplerate']), 'Back'],
			[1, 1, 1, 1, 1, 0],
			[doFullscreen, doMusic, doSound, doBuffer, doSamplerate, None]
		],
		[
			['Standard Levels','Custom Levels','Back'],
			[startGame, startCustomGame, 0],
			[None, None, None]
		],
		[
			[scores[0][0] + ' ' + str(scores[0][1]),scores[1][0] + ' ' + str(scores[1][1]),scores[2][0] + ' ' + str(scores[2][1]),scores[3][0] + ' ' + str(scores[3][1]),scores[4][0] + ' ' + str(scores[4][1]),scores[5][0] + ' ' + str(scores[5][1]),scores[6][0] + ' ' + str(scores[6][1]),scores[7][0] + ' ' + str(scores[7][1]),scores[8][0] + ' ' + str(scores[8][1]), scores[9][0] + ' ' + str(scores[9][1]), 'Back'],
			[3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 0],
			[None, None, None, None, None, None, None, None, None, None, None]
		]
	]
	return mainMenu

# render & interact with menus
class menu:
	def __init__(self, menuFunc, page = 0, option = 0):
		self.menu = menuFunc()
		self.menuFunc = menuFunc
		self.page = page
		self.option = option
		# load demo level
		self.level = pickle.load(open(os.path.join('levels','demo.map'),'r'))
		global font
		font.set_bold(True)
	def __del__(self):
		global font
		# prevent a 'NoneType' attribute error on game exit
		if isinstance(font, pygame.font.Font):
			font.set_bold(False)
	def draw(self, screen):
		global tileset
		global objectTileset
		drawBackground(self.level[0], tileset, screen) 
		drawObjectsAndMobs(self.level[1], objectTileset, [], screen, 0)
		# draw the menu itself
		global font
		optionImages = []
		counter = 0
		totalHeight = 0
		for option in self.menu[self.page][0]:
			if counter == self.option:
				colour = (255,0,0)
			else:
				colour = (0,0,128)
			counter += 1
			newImage = font.render(option,True,colour)
			totalHeight += newImage.get_height()
			optionImages.append(newImage)
		y = int(300 - (totalHeight / 2))
		for image in optionImages:
			screen.blit(image,(400 - int(image.get_width() / 2), y))
			y += image.get_height()
		# parse keypresses
		for key in pygame.event.get(KEYDOWN):
			if key.key == K_DOWN:
				self.option += 1
				if self.option == len(self.menu[self.page][0]):
					self.option = 0
			elif key.key == K_UP:
				self.option -= 1
				if self.option == -1:
					self.option = len(self.menu[self.page][0]) -1
			elif key.key == K_RETURN or key.key == K_ESCAPE:
				if key.key == K_RETURN:
					item = self.menu[self.page][1][self.option]
				else:
					item = self.menu[self.page][1][len(self.menu[self.page][1])-1]
				if isinstance(item,int):
					self.page = item
					self.option = 0
				else:
					item()
			elif key.key == K_LEFT or key.key == K_RIGHT:
				item = self.menu[self.page][2][self.option]
				if item != None:
					item(key.key == K_LEFT)
					self.menu = self.menuFunc()

##################
#                #
# Main game code #
#                #
##################

# enter name of a level set then start the game
class enterLevelName:
	def __init__(self):
		self.name = ''
	def draw(self,screen):
		global font
		global nextLoop
		# draw prompt
		screen.blit(font.render('Level set: ',True,(255,255,255)),(0,0))
		# draw filename
		screen.blit(font.render(self.name,True,(255,0,0)),(0,100))
		# edit filename
		for key in pygame.event.get(KEYDOWN):
			# allow a-z, 0-9 and underscores
			if (key.key > 96 and key.key < 123) or (key.key > 47 and key.key < 58) or key.key == 95:
				self.name += chr(key.key)
			# backspace to delete
			elif key.key == K_BACKSPACE:
				self.name = self.filename[:len(self.name)-1]
			# return to perform action
			elif key.key == K_RETURN:
				nextLoop = mainGame(self.name)
			# escape to cancel
			elif key.key == K_ESCAPE:
				nextLoop = menu(getMainMenu)

# tile groups (these are our "tetris blocks")
validTileGroups = [
	# straight
	[
		[[-1,-1,-1],[(0,0),(0,0),(0,0)],[-1,-1,-1]],
		[[-1,(0,1),-1],[-1,(0,1),-1],[-1,(0,1),-1]]
	],
	# L-shaped
	[
		[[-1,-1,(0,1)],[(0,0),(0,0),(0,5)],[-1,-1,-1]],
		[[(0,0),(0,3),-1],[-1,(0,1),-1],[-1,(0,1),-1]],
		[[-1,-1,-1],[(0,2),(0,0),(0,0)],[(0,1),-1,-1]],
		[[-1,(0,1),-1],[-1,(0,1),-1],[-1,(0,4),(0,0)]]
	],
	# flipped L
	[
		[[(0,1),-1,-1],[(0,4),(0,0),(0,0)],[-1,-1,-1]],
		[[-1,(0,1),-1],[-1,(0,1),-1],[(0,0),(0,5),-1]],
		[[-1,-1,-1],[(0,0),(0,0),(0,3)],[-1,-1,(0,1)]],
		[[-1,(0,2),(0,0)],[-1,(0,1),-1],[-1,(0,1),-1]]
	],
	# S-shaped
	[
		[[-1,-1,(0,1)],[(0,2),(0,0),(0,5)],[(0,1),-1,-1]],
		[[(0,0),(0,3),-1],[-1,(0,1),-1],[-1,(0,4),(0,0)]]
	],
	# flipped S
	[
		[[(0,1),-1,-1],[(0,4),(0,0),(0,3)],[-1,-1,(0,1)]],
		[[-1,(0,2),(0,0)],[-1,(0,1),-1],[(0,0),(0,5),-1]]
	],
	# C-shaped
	[
		[[-1,-1,-1],[(0,2),(0,0),(0,3)],[(0,1),-1,(0,1)]],
		[[-1,(0,2),(0,0)],[-1,(0,1),-1],[-1,(0,4),(0,0)]],
		[[(0,1),-1,(0,1)],[(0,4),(0,0),(0,5)],[-1,-1,-1]],
		[[(0,0),(0,3),-1],[-1,(0,1),-1],[(0,0),(0,5),-1]]
	],
	# corner piece
	[
		[[-1,-1,-1],[(0,0),(0,3),-1],[-1,(0,1),-1]],
		[[-1,-1,-1],[-1,(0,2),(0,0)],[-1,(0,1),-1]],
		[[-1,(0,1),-1],[-1,(0,4),(0,0)],[-1,-1,-1]],
		[[-1,(0,1),-1],[(0,0),(0,5),-1],[-1,-1,-1]]
	]
]

def inRange(x,y):
	if x > -1 and x < 20 and y > -1 and y < 20:
		return True
	else:
		return False

# main game loop
class mainGame:
	global validTileGroups
	global overlayTilesets
	global font
	global settings
	def __init__(self, set = ''):
		self.markerImage = makeImage(os.path.join('gfx','marker.bmp'))
		self.currentLevel = 0
		self.oldScore = -1
		self.oldLives = -1
		self.score = 0
		self.mode = 0
		self.lives = 4
		self.set = set
		self.gridHighlighter = makeImage(os.path.join('gfx','current_square.bmp'))
		font.set_bold(False)
		self.livesImage = font.render('Lives: ',True,(186,209,161))
		self.scoreImage = font.render('Score: ',True,(186,209,161))
		self.gameOverImage = font.render('Game Over',True,(0,0,0))
		self.frameno = 0
		self.levelComplete = font.render('Level Complete',True,(0,0,0))
		self.gameComplete = font.render('All Levels Complete!',True,(0,0,0))
		self.GOMusicPlaying = False
		if pygame.mixer.get_init() != None:
			if settings['sound']:
				self.soundOn = True
				self.newTrainSound = pygame.mixer.Sound(os.path.join('snd','newtrain.wav'))
				self.loseLifeSound = pygame.mixer.Sound(os.path.join('snd','loselife.wav'))
				self.gameOverSound = pygame.mixer.Sound(os.path.join('snd','gameover.wav'))
				self.thudSound = pygame.mixer.Sound(os.path.join('snd','thud.wav'))
			else:
				self.soundOn = False
			if settings['music']:
				self.musicOn = True
			else:
				self.musicOn = False
	def draw(self, screen):
		global currLoop
		global nextLoop
		self.frameno += 1
		if self.mode == 0:
			# load next level
			self.currentLevel += 1
			try:
				self.level = pickle.load(open(os.path.join('levels',self.set+str(self.currentLevel)+'.map'),'r'))
			except:
				self.mode = 5
				return
			# load level goals
			f = open(os.path.join('levels',self.set+str(self.currentLevel)+'.goals'),'r')
			i = 0
			self.targetSquares = []
			for line in f.readlines():
				line = line.rstrip()
				# ignore comments
				if line[0] == '#':
					continue
				elif i == 0:
					self.targetScore = int(line)
					self.targetScoreImage = font.render('Goal Score: '+line,True,(0,255,255))
					i = 1
				elif i == 1:
					self.targetTrains = int(line)
					self.targetTrainsImage = font.render('Goal Trains: '+line,True,(0,255,255))
					i = 2
				elif i == 2:
					self.timeLimit = int(line)
					self.timeRemaining = self.timeLimit
					self.timeLimitImage = font.render('Time: '+line,True,(0,255,255))
					i = 3
				elif i == 3:
					target = line.split(',')
					self.targetSquares.append([int(target[0]),int(target[1])])
			# blank initial overlay set & mobs
			self.mobs = []
			self.tracks = makeEmptyArray(-1)
			# progress to next state
			self.levelStart = font.render('Level '+str(self.currentLevel),True,(0,0,0))
			self.mode = 4
			# set initial current & next tiles
			self.currentTiles = None
			self.x = 9
			self.y = 9
			self.nextTiles = random.choice(validTileGroups)
			self.tileSpeed = 0.25
		elif self.mode == 1:
			# game in play!
			global objectTileset
			global tileset
			drawBackground(self.level[0],tileset,screen)
			drawOverlay(self.tracks,overlayTilesets,screen)
			drawObjectsAndMobs(self.level[1],objectTileset,self.mobs,screen,self.frameno,self.targetSquares)
			# remove invalid mobs
			for mymob in self.mobs:
				if not mymob.valid:
					self.mobs.remove(mymob)
					del mymob
					self.lives -= 1
					if self.lives >= 0 and self.soundOn:
						# play lose life sound, but only if not game over!
						self.loseLifeSound.play()
			# game over?
			if self.lives < 0:
				self.mode = 2
				if self.soundOn:
					self.gameOverSound.play()
			screen.blit(self.livesImage,(0,599-self.livesImage.get_height()))
			screen.blit(self.scoreImage,(0,0))
			if self.score != self.oldScore:
				self.oldScore = self.score
				self.pointsImage = font.render(str(self.score),True,(255,255,0))
			if self.lives != self.oldLives:
				self.oldLives = self.lives
				self.numlivesImage = font.render(str(self.lives),True,(255,0,0))
			screen.blit(self.numlivesImage,(self.livesImage.get_width(),599-self.livesImage.get_height()))
			screen.blit(self.pointsImage,(self.scoreImage.get_width(),0))
			# draw 'go thru' markers
			for target in self.targetSquares:
				screen.blit(self.markerImage,(381+(target[0]*18)-(target[1]*18),129+(target[1]*9)+(target[0]*9)-self.markerImage.get_height()))
			# pick new next tiles if necessary
			if self.currentTiles == None:
				self.currentTiles = self.nextTiles
				self.nextTiles = random.choice(validTileGroups)
				self.tileHeight = 100
				self.rotation = 0
			y = -1
			for currentRow in self.currentTiles[self.rotation]:
				x = -1
				for column in currentRow:
					if column != -1:
						gx = 381 + ((self.x+x)*18) - ((self.y+y)*18)
						gy = 111 + ((self.x+x)*9) + ((self.y+y)*9)
						# draw highlighter
						screen.blit(self.gridHighlighter,(gx,gy))
						# draw current tile highlighter
						screen.blit(self.gridHighlighter,(gx, gy - int(self.tileHeight)))
						# and current tile
						screen.blit(overlayTilesets[column[0]].tiles[column[1]].image,(gx, gy - int(self.tileHeight)))
					x += 1
				y += 1
					
			# process keypresses
			for key in pygame.event.get(KEYDOWN):
				if key.key == K_DOWN and self.y < 19:
					self.y += 1
				elif key.key == K_UP and self.y > 0:
					self.y -= 1
				elif key.key == K_LEFT and self.x > 0:
					self.x -= 1
				elif key.key == K_RIGHT and self.x < 19:
					self.x += 1
				elif key.key == K_SPACE or key.key == K_RETURN:
					# drop tile so it lands this turn
					self.tileHeight = self.tileSpeed
				elif key.key == K_PAGEUP:
					self.rotation += 1
					if self.rotation == len(self.currentTiles):
						self.rotation = 0
				elif key.key == K_PAGEDOWN:
					self.rotation -= 1
					if self.rotation == -1:
						self.rotation = len(self.currentTiles)-1
			# move current tile down
			self.tileHeight -= self.tileSpeed
			if self.tileHeight == 0:
				if self.soundOn:
					self.thudSound.play()
				# lay all the tiles, keeping track of how many cannot be lain
				numPiecesNotLain = 0
				sy = self.y - 1
				for currentRow in self.currentTiles[self.rotation]:
					sx = self.x - 1
					for column in currentRow:
						# if the ground is clear, lay the tile
						if column != -1 and inRange(sy,sx) and self.level[0][sy][sx] == 0 and self.level[1][sy][sx] == -1:
							self.tracks[sy][sx] = (column[0], column[1])
						elif column != -1:
							numPiecesNotLain += 1
						sx += 1
					sy += 1
				# check for track loops
				sy = self.y - 1
				for currentRow in self.currentTiles[self.rotation]:
					sx = self.x - 1
					for column in currentRow:
						if column != -1 and inRange(sy,sx) and self.tracks[sy][sx] != -1:
							# check our endpoints and follow the track to see if there is a valid track loop
							x = sx
							y = sy
							oldx = x
							oldy = y
							valid = True
							done = False
							source = overlayTilesets[column[0]].tiles[column[1]]
							# keep track of how many pieces are in this loop (if it does turn out to be a loop)
							numPieces = 0
							while valid:
								valid = False
								for route in source.routingTable:
									# check outgoing routes from current square, preventing backtracking
									if route[1] == '0' and y > 0 and self.tracks[y-1][x] != -1 and (x,y-1) != (oldx,oldy):
										dx = 0
										dy = -1
									elif route[1] == '1' and x > 0 and self.tracks[y][x-1] != -1 and (x-1,y) != (oldx,oldy):
										dx = -1
										dy = 0
									elif route[1] == '2' and x < 19 and self.tracks[y][x+1] != -1 and (x+1,y) != (oldx,oldy):
										dx = 1
										dy = 0
									elif route[1] == '3' and y < 19 and self.tracks[y+1][x] != -1 and (x,y+1) != (oldx,oldy):
										dx = 0
										dy = 1
									else:
										continue
									dest = overlayTilesets[self.tracks[y+dy][x+dx][0]].tiles[self.tracks[y+dy][x+dx][1]]
									for secondroute in dest.routingTable:
										# check incoming routes from destination square
										if sourceMap[route[1]] == secondroute[0]:
											oldx = x
											oldy = y
											x += dx
											y += dy
											# if there is not already a mob here...
											mobAlreadyThere = False
											for mymob in self.mobs:
												if mymob.location[0] == x and mymob.location[1] == y:
													mobAlreadyThere = True
													break
											if not mobAlreadyThere:
												# we have a next step
												valid = True
												numPieces += 1
												source = dest
												# do we have a loop?
												if x == sx and y == sy:
													done = True
											break
									break
								if done:
									break
							if done:
								global trainset
								self.mobs.append(mob(trainset,(x,y),self.tracks,overlayTilesets))
								# add points
								self.score += numPieces * numPieces
								self.pointsThisLevel += numPieces * numPieces
								if self.soundOn:
									self.newTrainSound.play()
						sx += 1
					sy += 1
				self.currentTiles = None
				# subtract points for pieces not lain
				self.score -= numPiecesNotLain * 20
			# check level goals (and render the goal markers):
			complete = True
			y = 599
			if self.targetScore != 0:
				y -= self.targetScoreImage.get_height()
				screen.blit(self.targetScoreImage,(799-self.targetScoreImage.get_width(),y))
				if self.pointsThisLevel != self.targetScore:
					complete = False
			if self.targetTrains != 0:
				y -= self.targetTrainsImage.get_height()
				screen.blit(self.targetTrainsImage,(799-self.targetTrainsImage.get_width(),y))
				if len(self.mobs) != self.targetTrains:
					complete = False
			if self.timeLimit != 0:
				# reduce the time remaining
				for event in pygame.event.get(USEREVENT):
					self.timeRemaining -= 1
					self.timeLimitImage = font.render('Time: '+str(self.timeRemaining),True,(0,255,255))
				y -= self.timeLimitImage.get_height()
				screen.blit(self.timeLimitImage,(799-self.timeLimitImage.get_width(),y))
				if self.timeRemaining == 0:
					complete = False
					# when time limit is up, lose a life and restart the level
					self.lives -= 1
					self.mode = 0
					self.currentLevel -= 1
			if len(self.targetSquares) != 0:
				complete = False
				for mymob in self.mobs:
					hitTargets = True
					for targetSquare in self.targetSquares:
						#print "checking "+str(targetSquare)+" against "+str(mymob.targetsHit)
						if mymob.targetsHit == None or (targetSquare not in mymob.targetsHit):
							hitTargets = False
							break
					if hitTargets:
						complete = True
						break
			if complete:
				self.mode = 3
		elif self.mode == 2:
			if self.musicOn and not self.GOMusicPlaying:
				pygame.mixer.music.load(os.path.join('snd','gameover.mod'))
				pygame.mixer.music.play(-1)
				self.GOMusicPlaying = True
			# game over screen
			pygame.time.set_timer(USEREVENT,0)
			drawBackground(self.level[0],tileset,screen)
			drawOverlay(self.tracks,overlayTilesets,screen)
			drawObjectsAndMobs(self.level[1],objectTileset,self.mobs,screen,self.frameno)
			for mymob in self.mobs:
				if not mymob.valid:
					self.mobs.remove(mymob)
					del mymob
			screen.blit(self.gameOverImage,(400-int(self.gameOverImage.get_width()/2),300-int(self.gameOverImage.get_height()/2)))
			gone = False
			for key in pygame.event.get(KEYDOWN):
				gone = True
			if gone:
				currLoop = None
				nextLoop = highScoreEntry(self.score)
				self.GOMusicPlaying = False
		elif self.mode == 3:
			# level complete screen
			pygame.time.set_timer(USEREVENT,0)
			drawBackground(self.level[0],tileset,screen)
			drawOverlay(self.tracks,overlayTilesets,screen)
			drawObjectsAndMobs(self.level[1],objectTileset,self.mobs,screen,self.frameno)
			for mymob in self.mobs:
				if not mymob.valid:
					self.mobs.remove(mymob)
					del mymob
			screen.blit(self.levelComplete,(400-int(self.levelComplete.get_width()/2),300-int(self.levelComplete.get_height()/2)))
			gone = False
			for key in pygame.event.get(KEYDOWN):
				gone = True
			if gone:
				self.mode = 0
		elif self.mode == 4:
			# level start screen
			pygame.time.set_timer(USEREVENT,0)
			if self.timeLimit != 0:
				# todo: move this to the keypress handler in level message display state
				pygame.time.set_timer(USEREVENT,1000)
			drawBackground(self.level[0],tileset,screen)
			drawOverlay(self.tracks,overlayTilesets,screen)
			drawObjectsAndMobs(self.level[1],objectTileset,self.mobs,screen,self.frameno)
			screen.blit(self.levelStart,(400-int(self.levelStart.get_width()/2),300-int(self.levelStart.get_height()/2)))
			gone = False
			for key in pygame.event.get(KEYDOWN):
				gone = True
			if gone:
				self.mode = 1
				if self.musicOn:
					pygame.mixer.music.load(os.path.join('snd','ingame.it'))
					pygame.mixer.music.play(-1)
			self.pointsThisLevel = 0
		elif self.mode == 5:
			# game complete screen
			pygame.time.set_timer(USEREVENT,0)
			drawBackground(self.level[0],tileset,screen)
			drawOverlay(self.tracks,overlayTilesets,screen)
			drawObjectsAndMobs(self.level[1],objectTileset,self.mobs,screen,self.frameno)
			for mymob in self.mobs:
				if not mymob.valid:
					self.mobs.remove(mymob)
					del mymob
			screen.blit(self.gameComplete,(400-int(self.gameComplete.get_width()/2),300-int(self.gameComplete.get_height()/2)))
			gone = False
			for key in pygame.event.get(KEYDOWN):
				gone = True
			if gone:
				currLoop = None
				nextLoop = highScoreEntry(self.score)

# high score entry code
class highScoreEntry:
	def __init__(self, points):
		global scores
		global font
		# see if score qualifies
		counter = 0
		found = False
		for score in scores:
			if points >= score[1]:
				found = True
				break
			counter += 1
		if not found:
			global currLoop
			global nextLoop
			currLoop = None
			nextLoop = main(getMainMenu)
			self.valid = False
		else:
			# it does
			self.valid = True
			self.points = points
			self.position = counter
			self.name = ''
			if settings['music'] and pygame.mixer.get_init() != None:
				pygame.mixer.music.load(os.path.join('snd','hiscores.mod'))
				pygame.mixer.music.play(-1)
			self.scoresImage = font.render('Enter your name:',True,(0,255,0))
	def draw(self,screen):
		global font
		global scores
		if not self.valid:
			return
		screen.blit(self.scoresImage,(0,0))
		for key in pygame.event.get(KEYDOWN):
			# allow a-z, 0-9
			if (key.key > 96 and key.key < 123) or (key.key > 47 and key.key < 58):
				mods = pygame.key.get_mods()
				caps = False
				# allow capital letters also
				if (key.key > 96 and key.key < 123):
					if mods & KMOD_SHIFT:
						caps = True
					if mods & KMOD_CAPS:
						if caps:
							caps = False
						else:
							caps = True
				char = chr(key.key)
				if caps:
					char = chr(key.key - 32)
				self.name += char
			# backspace to delete
			elif key.key == K_BACKSPACE:
				self.name = self.name[:len(self.name)-1]
			# return to perform action
			elif key.key == K_RETURN:
				for i in range(8,self.position-1,-1):
					scores[i+1][0] = scores[i][0]
					scores[i+1][1] = scores[i][1]
				scores[self.position][0] = self.name
				scores[self.position][1] = self.points
				global currLoop
				global nextLoop
				currLoop = None
				# todo: pass in initial page & option, to highlight new entry
				nextLoop = menu(getMainMenu,3,self.position)
		screen.blit(font.render(self.name,True,(255,0,0)),(0,100))

#####################
#                   #
# Level editor code #
#                   #
#####################

class levelEditor:
	def __init__(self):
		global tileset
		self.level = []
		# background
		self.level.append(makeEmptyArray(0))
		# overlays (no starting overlays in levels)
		#self.level.append(makeEmptyArray(-1))
		# objects (no starting mobs in levels)
		self.level.append(makeEmptyArray(-1))
		# filename
		self.filename = ''
		# edit cursor position
		self.x = 0
		self.y = 0
		# tile set index & current tile set (ground, overlay, object)
		self.setIndex = 0
		self.set = 0
		# highlighting graphics
		# selected tile type
		self.tileHighlighter = makeImage(os.path.join('gfx','current_tile.bmp'))
		# selected grid square
		self.gridHighlighter = makeImage(os.path.join('gfx','current_square.bmp'))
		# mode = map editor
		self.mode = 0
	def draw(self, screen):
		if self.mode == 0:
			# draw level as it currently stands
			global objectTileset
			global tileset
			drawBackground(self.level[0],tileset,screen)
			drawObjectsAndMobs(self.level[1], objectTileset, [], screen, 0)
			# draw current tile set as a row across the top
			if self.set == 0:
				tiles = tileset
			else:
				tiles = objectTileset
			count = 0
			for tile in tiles:
				screen.blit(tile,(381 - ((self.setIndex - count) * 38), 0))
				count += 1
			# draw editing highlighters
			screen.blit(self.tileHighlighter,(381,0))
			screen.blit(self.gridHighlighter,(381 + (self.x*18) - (self.y*18), 111 + (self.x*9) + (self.y*9)))
			# process keyboard input
			for key in pygame.event.get(KEYDOWN):
				if pygame.key.get_mods() & KMOD_SHIFT:
					# shift & left/right scrolls through editable tiles
					if key.key == K_RIGHT and self.setIndex < (len(tiles)-1):
						self.setIndex += 1
					elif key.key == K_LEFT and self.setIndex > 0:
						self.setIndex -= 1
				else:
					if key.key == K_UP and self.y > 0:
						self.y -= 1
					elif key.key == K_DOWN and self.y < (len(self.level[0])-1):
						self.y += 1
					elif key.key == K_RIGHT and self.x < (len(self.level[0][0])-1):
						self.x += 1
					elif key.key == K_LEFT and self.x > 0:
						self.x -= 1
					elif key.key == K_RETURN:
						self.level[self.set][self.y][self.x] = self.setIndex
					elif key.key == K_BACKSPACE or key.key == K_DELETE:
						self.level[1][self.y][self.x] = -1
					elif key.key == K_1 or key.key == K_KP1:
						# change tileset (ground)
						self.set = 0
						self.setIndex = 0
					elif key.key == K_2 or key.key == K_KP2:
						# change tileset (objects)
						self.set = 1
						self.setIndex = 0
					elif key.key == K_l:
						# load
						self.filename = ''
						self.mode = 1
					elif key.key == K_s:
						# save
						self.mode = 2
					elif key.key == K_n:
						# new blank level
						self.filename = ''
						self.level = []
						self.level.append(makeEmptyArray(0))
						self.level.append(makeEmptyArray(-1))
					elif key.key == K_ESCAPE:
						# back to main menu
						global nextLoop
						nextLoop = menu(getMainMenu)
		elif self.mode == 1 or self.mode == 2:
			# load or save
			global font
			# draw prompt
			if self.mode == 1:
				screen.blit(font.render('Filename (load): ',True,(255,255,255)),(0,0))
			else:
				screen.blit(font.render('Filename (save): ',True,(255,255,255)),(0,0))
			# draw filename
			screen.blit(font.render(self.filename,True,(255,0,0)),(0,100))
			# edit filename
			for key in pygame.event.get(KEYDOWN):
				# allow a-z, 0-9 and underscores
				if (key.key > 96 and key.key < 123) or (key.key > 47 and key.key < 58) or key.key == 95:
					self.filename += chr(key.key)
				# backspace to delete
				elif key.key == K_BACKSPACE:
					self.filename = self.filename[:len(self.filename)-1]
				# return to perform action
				elif key.key == K_RETURN:
					if self.mode == 1:
						self.level = pickle.load(open(os.path.join('levels',self.filename+'.map'),'r'))
					else:
						pickle.dump(self.level,open(os.path.join('levels',self.filename+'.map'),'w'))
					self.set = 0
					self.setIndex = 0
					self.x = 0
					self.y = 0
					self.mode = 0
				# escape to cancel
				elif key.key == K_ESCAPE:
					self.mode = 0

#######################
#                     #
# Program entry point #
#                     #
#######################

random.seed()

try:
	pygame.display.init()
except Exception, e:
	sys.exit('Could not initialise display module: '.str(e))

soundWarning = False
if settings['sound'] or settings['music']:
	try:
		pygame.mixer.pre_init(settings['samplerate'], -16, 16, settings['buffer'])
		pygame.mixer.init()
	except Exception, e:
		soundWarning = True
		settings['sound'] = False

pygame.font.init()
font = pygame.font.Font(os.path.join('gfx','VeraSe.ttf'),28)

# hide the pointer
pygame.mouse.set_visible(False)

# set screen mode
screenmode = HWSURFACE | DOUBLEBUF
if settings['fullscreen']:
	screenmode |= FULLSCREEN
screen = pygame.display.set_mode((800, 600), screenmode)

# TODO:
# Make a 'levels' subdirectory, which contains an 'index' manifest (level sets, names etc.), and subdirectories containing the levels themselves.
# Level files need some form of goal description, plus arrays for ground cubes, starting overlays, and objects. Goals could include time limits (survival),
# scores, making tracks that go through certain squares, making a certain number of tracks, etc. Have multiple goals with AND/OR for completion.
#
# Object manifest will need to define whether or not object is 'special': get a score for having a valid track running past it, for example.
#
# Make a global manifest, detailing all ground tiles, mobs, etc. in terms of the various sub-manifests, instead of referring to them manually?
# For now, hard-code the latter, as being flexible in defining what mobs run on what paths, what pathset the player is given tiles from, etc. is out of scope for one week's work.

# load ground tiles
tileset = makeGroundTiles('ground')

# load path overlays
overlayTilesets = []
overlayTilesets.append(pathTileSet('rails',0))

# load mobs
mobs = []
trainset = loadMobTiles('train')

# load objects
objectTileset = makeObjectTileSet('objects')

def makeEmptyArray(v):
	array = []
	for i in range(20):
		array2 = []
		for j in range(20):
			array2.append(v)
		array.append(array2)
		del array2
	return array

# make default level
level = []
# background
level.append(makeEmptyArray(0))
# overlay
level.append(makeEmptyArray(-1))

#mobs.append(mob(trainset,(9,14),level[1],overlayTilesets))

# block all events, clear the queue, then listen for events we're interested in
pygame.event.set_allowed(None)
pygame.event.clear()
pygame.event.set_allowed([QUIT, KEYDOWN, USEREVENT])

# enable key repeat
pygame.key.set_repeat(300,30)

# first render/interaction loop to run
nextLoop = None
currLoop = menu(getMainMenu)

while True:
	screen.fill((0,0,0))
	if nextLoop != None:
		pygame.event.clear()
		pygame.time.set_timer(USEREVENT,0)
		del currLoop
		currLoop = nextLoop
		del nextLoop
		nextLoop = None
		if settings['music'] and pygame.mixer.get_init() != None and pygame.mixer.music.get_busy():
			pygame.mixer.music.stop()
	currLoop.draw(screen)
	pygame.display.flip()
	pygame.time.wait(30)
	#pygame.event.pump()
	if pygame.event.peek(QUIT):
		break
	#pygame.event.clear()

sys.exit()
